/**
 * An efficient implementation of container for logging read and
 * write sets (and possibly other things).
 * 
 * This log can be used as a memory pool for logging word values too.
 *
 * @author Aleksandar Dragojevic aleksandar.dragojevic@epfl.ch
 */

#ifndef WLPDSTM_ARRAY_H_
#define WLPDSTM_ARRAY_H_

#include "../allocators.h"

#include <cstring>

#ifdef CHUNKED_LOG

namespace wlpdstm {

	template<typename T, int CHUNK_LENGTH = 2048>
	class Log : public WlpdstmAlloced {
		private:
			struct LogChunk : public WlpdstmAlloced {
				LogChunk() : next(NULL), prev(NULL) { }

				T elements[CHUNK_LENGTH];
				LogChunk *next;
				LogChunk *prev;
			};

		public:
			class iterator : public WlpdstmAlloced {
				public:
					iterator(LogChunk *c = NULL, unsigned el = 0,
							LogChunk *lc = NULL, unsigned lel = 0)
						: chunk(c), element(el), lastChunk(lc),
						lastElement(lel) { }

				public:
					bool operator==(const iterator& rhs);
					bool operator!=(const iterator& rhs);
					void next();
					bool hasNext();
					iterator& operator++();
					iterator operator++(int);
					T& operator*() const;

				private:
					LogChunk *chunk;
					unsigned element;
					LogChunk *lastChunk;
					unsigned lastElement;
			};

		public:
			Log() : head(new LogChunk()), lastChunk(head),
				nextElement(0) { }

			iterator begin() const;
			iterator end() const;

			bool empty() const;
			void insert(const T& element);
			T remove();
			void delete_last();
			T *get_next();
			void clear();

			bool contains(const T *el);

		private:
			LogChunk *head;
			// logically last chunk
			LogChunk *lastChunk;
			unsigned nextElement;
	};
}

namespace wlpdstm {
	
	template<typename T, int CHUNK_LENGTH>
	inline typename Log<T, CHUNK_LENGTH>::iterator Log<T, CHUNK_LENGTH>::begin() const {
		return iterator(head, 0, lastChunk, nextElement);
	}

	template<typename T, int CHUNK_LENGTH>
	inline typename Log<T, CHUNK_LENGTH>::iterator Log<T, CHUNK_LENGTH>::end() const {
		return iterator(lastChunk, nextElement, lastChunk, nextElement);
	}

	template<typename T, int CHUNK_LENGTH>
	inline bool Log<T, CHUNK_LENGTH>::empty() const {
		return (nextElement == 0 && lastChunk == head);
	}

	template<typename T, int CHUNK_LENGTH>
	inline void Log<T, CHUNK_LENGTH>::insert(const T& el) {
		lastChunk->elements[nextElement++] = el;

		if(nextElement == CHUNK_LENGTH) {
			LogChunk *chunk = lastChunk->next;

			if(chunk == NULL) {
				chunk = new LogChunk();
				lastChunk->next = chunk;
				chunk->prev = lastChunk;
			}

			lastChunk = chunk;
			nextElement = 0;
		}
	}

	template<typename T, int CHUNK_LENGTH>
	inline T *Log<T, CHUNK_LENGTH>::get_next() {
		T *ret = lastChunk->elements + nextElement++;
		//lastChunk->elements[nextElement++] = el;

		if(nextElement == CHUNK_LENGTH) {
			LogChunk *chunk = lastChunk->next;

			if(chunk == NULL) {
				chunk = new LogChunk();
				lastChunk->next = chunk;
				chunk->prev = lastChunk;
			}

			lastChunk = chunk;
			nextElement = 0;
		}

		return ret;
	}

	template<typename T, int CHUNK_LENGTH>
	inline bool Log<T, CHUNK_LENGTH>::contains(const T *el) {
		LogChunk *curr = head;

		while(curr != NULL) {
			const T *chunk_start = curr->elements;
			const T *chunk_end = chunk_start + CHUNK_LENGTH - 1;

			if(el >= chunk_start && el <= chunk_end) {
				return true;
			}

			if(curr == lastChunk) {
				return false;
			}

			curr = curr->next;
		}

		return false;
	}

	template<typename T, int CHUNK_LENGTH>
	inline T Log<T, CHUNK_LENGTH>::remove() {
		if(nextElement == 0) {
			nextElement = CHUNK_LENGTH - 1;
			lastChunk = lastChunk->prev;
			return lastChunk->elements[CHUNK_LENGTH - 1];
		}

		return lastChunk->elements[--nextElement];
	}

	template<typename T, int CHUNK_LENGTH>
	inline void Log<T, CHUNK_LENGTH>::delete_last() {
		if(nextElement == 0) {
			nextElement = CHUNK_LENGTH - 1;
			lastChunk = lastChunk->prev;
		} else {
			--nextElement;
		}
	}

	template<typename T, int CHUNK_LENGTH>
	inline void Log<T, CHUNK_LENGTH>::clear() {
		lastChunk = head;
		nextElement = 0;
	}

	template<typename T, int CHUNK_LENGTH>
	inline bool Log<T, CHUNK_LENGTH>::iterator::operator==(
			const Log<T, CHUNK_LENGTH>::iterator::iterator& rhs) {
		return chunk == rhs.chunk && element == rhs.element;
	}

	template<typename T, int CHUNK_LENGTH>
	inline bool Log<T, CHUNK_LENGTH>::iterator::operator!=(
			const Log<T, CHUNK_LENGTH>::iterator::iterator& rhs) {
		return !(chunk == rhs.chunk && element == rhs.element);
	}

	template<typename T, int CHUNK_LENGTH>
	inline void Log<T, CHUNK_LENGTH>::iterator::next() {
		element++;

		if(element == CHUNK_LENGTH) {
			element = 0;
			chunk = chunk->next;
		}
	}

	template<typename T, int CHUNK_LENGTH>
	inline bool Log<T, CHUNK_LENGTH>::iterator::hasNext() {
		return !(element == lastElement && chunk == lastChunk);
	}

	template<typename T, int CHUNK_LENGTH>
	inline typename Log<T, CHUNK_LENGTH>::iterator& Log<T, CHUNK_LENGTH>::iterator::operator++() {
		next();
		return this;
	}

	template<typename T, int CHUNK_LENGTH>
	inline typename Log<T, CHUNK_LENGTH>::iterator Log<T, CHUNK_LENGTH>::iterator::operator++(int dummy) {
		typename Log<T, CHUNK_LENGTH>::iterator ret = *this;
		next();
		return ret;
	}

	template<typename T, int CHUNK_LENGTH>
	inline T& Log<T, CHUNK_LENGTH>::iterator::operator*() const {
		return chunk->elements[element];
	}
}

#endif // chunked log


#ifdef VECTOR_LOG

namespace wlpdstm {

	class TxMixInv;

	template<typename T, int INITIAL_SIZE = 256>
	class Log : public WlpdstmAlloced {
		public:
			class iterator : public WlpdstmAlloced {
				public:
					iterator(T *ar = NULL, unsigned s = 0,
							unsigned c = 0)
						: array(ar), size(s), current(c) {  }

				public:
					bool operator==(const iterator& rhs);
					bool operator!=(const iterator& rhs);
					void next();
					bool hasNext();
					iterator& operator++();
					iterator operator++(int);
					T& operator*() const;

				private:
					T *array;
					unsigned size;
					unsigned current;
			};

		public:
			Log() : array((T *)wlpdstm::malloc(INITIAL_SIZE * sizeof(T))),
				size(INITIAL_SIZE), next(0) { }

			iterator begin() const;
			iterator end() const;

			bool empty() const;
			void insert(const T& element);
			void clear();
			T *get_next();

		private:
			T *array;
			unsigned size;
			unsigned next;
	};
}

namespace wlpdstm {

	template<typename T, int INITIAL_SIZE>
	inline typename Log<T, INITIAL_SIZE>::iterator Log<T, INITIAL_SIZE>::begin() const {
		return iterator(array, next);
	}

	template<typename T, int INITIAL_SIZE>
	inline typename Log<T, INITIAL_SIZE>::iterator Log<T, INITIAL_SIZE>::end() const {
		return iterator(array, next, next);
	}

	template<typename T, int INITIAL_SIZE>
	inline bool Log<T, INITIAL_SIZE>::empty() const {
		return (next == 0);
	}

	template<typename T, int INITIAL_SIZE>
	inline void Log<T, INITIAL_SIZE>::insert(const T& el) {
		if(next == size) {
			size *= 2;
			T *oldArray = array;
			array = (T *)wlpdstm::malloc(size * sizeof(T));
			memcpy(array, oldArray, next);
			wlpdstm::free(oldArray);
		}

		array[next++] = el;
	}

	template<typename T, int INITIAL_SIZE>
	inline void Log<T, INITIAL_SIZE>::clear() {
		next = 0;
	}

	template<typename T, int INITIAL_SIZE>
	inline bool Log<T, INITIAL_SIZE>::iterator::operator==(
			const Log<T, INITIAL_SIZE>::iterator::iterator& rhs) {
		return array == rhs.array && current == rhs.current;
	}

	template<typename T, int INITIAL_SIZE>
	inline bool Log<T, INITIAL_SIZE>::iterator::operator!=(
			const Log<T, INITIAL_SIZE>::iterator::iterator& rhs) {
		return !(array == rhs.array && current == rhs.current);
	}

	template<typename T, int INITIAL_SIZE>
	inline void Log<T, INITIAL_SIZE>::iterator::next() {
		current++;
	}

	template<typename T, int INITIAL_SIZE>
	inline bool Log<T, INITIAL_SIZE>::iterator::hasNext() {
		return (current < size) && (size != 0);
	}

	template<typename T, int INITIAL_SIZE>
	inline typename Log<T, INITIAL_SIZE>::iterator& Log<T, INITIAL_SIZE>::iterator::operator++() {
		next();
		return this;
	}

	template<typename T, int INITIAL_SIZE>
	inline typename Log<T, INITIAL_SIZE>::iterator Log<T, INITIAL_SIZE>::iterator::operator++(int dummy) {
		typename Log<T, INITIAL_SIZE>::iterator ret = *this;
		next();
		return ret;
	}

	template<typename T, int INITIAL_SIZE>
	inline T& Log<T, INITIAL_SIZE>::iterator::operator*() const {
		return array[current];
	}

	template<typename T, int INITIAL_SIZE>
	inline T *Log<T, INITIAL_SIZE>::get_next() {
		return array + next++;
	}

	template<typename T, int CHUNK_LENGTH>
	inline bool Log<T, CHUNK_LENGTH>::contains(const T *el) {
		return el >= array && el <= array + size - 1;
	}
}

#endif

#endif // WLPDSTM_ARRAY_H_
