/**
 * @author Aleksandar Dragojevic aleksandar.dragojevic@epfl.ch
 */

#ifndef WLPDSTM_MEMORY_H_
#define WLPDSTM_MEMORY_H_

#include <cstdlib>

#include "../common/atomic.h"
#include "../constants.h"

//////////////////////////////
// include allocators start //
//////////////////////////////

#ifdef MM_TBB_MALLOC
#include "scalable_allocator.h"
#elif defined MM_HOARD
#include "mm_hoard.h"
#elif defined MM_LEA
#include "mm_lea.h"
#endif

////////////////////////////
// include allocators end //
////////////////////////////

namespace wlpdstm {

	////////////////////
	// helper classes //
	////////////////////

	/**
	 * Vector timestamp that is useful for tracking when some object
	 * become private.
	 */
	class VectorTimestamp {
		friend 	bool operator<(const VectorTimestamp &older,
			const VectorTimestamp &newer);

		public:
			VectorTimestamp() {
				// do nothing
			}

			VectorTimestamp(const VectorTimestamp &vt);

			void Increment(unsigned tid);

			void RegisterThread(unsigned tid);

			void Init();

		private:
			unsigned long operator[](unsigned idx) const;

			unsigned long &GetElement(unsigned idx);

		private:
			unsigned long ts[wlpdstm::CACHE_LINE_SIZE * MAX_THREADS];

			unsigned length;
	} __attribute__ ((aligned(64)));

	bool operator<(const VectorTimestamp &older,
		const VectorTimestamp &newer);


	/**
	 * Struct used for buffering freed pointers.
	 */
	template<unsigned SIZE = 2048>
	struct PtrBufferT {
		// next to acquire
		unsigned next_ptr;

		// where to rollback
		unsigned rollback_next_ptr;

		// vector timestamp
		VectorTimestamp ts;

		bool init_ts;

		// pointers for chaining
		PtrBufferT<SIZE> *next;

		PtrBufferT<SIZE> *prev;

		void *ptrs[SIZE];

		void Clear() {
			next_ptr = 0;
			rollback_next_ptr = 0;
			next = NULL;
			prev = NULL;
			init_ts = false;
		}

		void Rollback() {
			next_ptr = rollback_next_ptr;
		}

		void Commit() {
			rollback_next_ptr = next_ptr;
		}

		bool IsFull() const {
			return next_ptr == SIZE;
		}

		unsigned GetSize() const {
			return SIZE;
		}
	};

	typedef PtrBufferT<> PtrBuffer;

	////////////////////
	// memory manager //
	////////////////////

	/**
	 * This is a main memory manager class. Typically, one instance of
	 * this class should be created per thread.
	 */
	class MemoryManager {
		public:
			// tx lifetime
			void TxStart();

			void TxCommit();

			void TxAbort();

			// allocation / deallocation
			static void *Malloc(size_t size);

			static void Free(void *ptr);

			void *TxMalloc(size_t size);

			void TxFree(void *ptr);

			// init functions
			static void GlobalInit();

			void ThreadInit(unsigned tid);

		private:
			PtrBuffer *AllocatePtrBuffer();

			PtrBuffer *GetCleanPtrBuffer();

			void PutReusable(PtrBuffer *first, PtrBuffer *last);

			void PutFreedLogically(PtrBuffer *first, PtrBuffer *last);

			void FreeDominated();

			void FreePtrs(PtrBuffer *buf);

			void FreeMemoryPtr(void *ptr);

		private:
			// cached thread id
			unsigned tid;

			// mem dealloced in this tx
			PtrBuffer *freed_in_tx;
			PtrBuffer *freed_in_tx_old;

			// mem allocated in this tx
			PtrBuffer *alloced_in_tx;
			PtrBuffer *alloced_in_tx_old;

			// logically freed pointers
			PtrBuffer *freed_logically;

			// reusable pointer buffers
			PtrBuffer *reusable;

			// global vector timestamp
			static VectorTimestamp mm_ts;
	};
}

/////////////////////////////////////////
// memory manager implementation start //
/////////////////////////////////////////

inline void wlpdstm::MemoryManager::GlobalInit() {
	mm_ts.Init();
}

inline void wlpdstm::MemoryManager::ThreadInit(unsigned t) {
	tid = t;
	
	// initialize buffer pointers
	freed_in_tx = AllocatePtrBuffer();
	freed_in_tx_old = freed_in_tx;

	alloced_in_tx = AllocatePtrBuffer();
	alloced_in_tx_old = alloced_in_tx;

	freed_logically = NULL;

	reusable = NULL;

	// update shared vector timestamp
	mm_ts.RegisterThread(tid);
}

inline void wlpdstm::MemoryManager::TxStart() {
	mm_ts.Increment(tid);
}

inline void wlpdstm::MemoryManager::TxCommit() {
	// first increment global timestamp
	mm_ts.Increment(tid);

	// forget allocations in this tx
	if(alloced_in_tx_old != alloced_in_tx) {
		PutReusable(alloced_in_tx_old->next, alloced_in_tx);

		// update old ptr
		alloced_in_tx = alloced_in_tx_old;
	}

	alloced_in_tx->next_ptr = 0;

	// store deletions in freed_logically
	freed_in_tx->Commit();

	if(freed_in_tx_old != freed_in_tx) {
		// copy timestamp to first pointer buffer
		freed_in_tx_old->ts = mm_ts;
		freed_in_tx_old->init_ts = true;

		// chain list at the beginning of freed_logically
		PutFreedLogically(freed_in_tx_old, freed_in_tx->prev);

		// update old ptr
		freed_in_tx_old = freed_in_tx;

		// try to actually release some memory
		FreeDominated();
	}
}

inline void wlpdstm::MemoryManager::TxAbort() {
	// first increment global timestamp
	mm_ts.Increment(tid);

	// forget deallocations in this tx
	freed_in_tx_old->Rollback();

	if(freed_in_tx_old != freed_in_tx) {
		PutReusable(freed_in_tx_old->next, freed_in_tx);

		// update old ptr
		freed_in_tx = freed_in_tx_old;
	}

	// rollback allocations
	// TODO: perhaps postpone this as well? although it does not seem
	// to make too much sense or buy anything
	for(PtrBuffer *curr = alloced_in_tx_old;curr != NULL;curr = curr->next) {
		FreePtrs(curr);
	}

	// put reusable
	if(alloced_in_tx != alloced_in_tx_old) {
		PutReusable(alloced_in_tx_old->next, alloced_in_tx);
		alloced_in_tx = alloced_in_tx_old;
	}

	// maintain pointers
	alloced_in_tx->next_ptr = 0;
}

inline wlpdstm::PtrBuffer *wlpdstm::MemoryManager::AllocatePtrBuffer() {
	PtrBuffer *ret = (PtrBuffer *)Malloc(sizeof(PtrBuffer));
	ret->Clear();
	return ret;
}

inline wlpdstm::PtrBuffer *wlpdstm::MemoryManager::GetCleanPtrBuffer() {
	if(reusable != NULL) {
		PtrBuffer *ret = reusable;
		PtrBuffer *next = reusable->next;
		reusable = next;

		if(next != NULL) {
			next->prev = NULL;
		}

		ret->Clear();

		return ret;
	}

	return AllocatePtrBuffer();
}

inline void wlpdstm::MemoryManager::PutReusable(
		PtrBuffer *first, PtrBuffer *last) {
	last->next = reusable;

	if(reusable != NULL) {
		reusable->prev = last;
	}

	first->prev = NULL;
	reusable = first;
}

inline void wlpdstm::MemoryManager::PutFreedLogically(
		PtrBuffer *first, PtrBuffer *last) {
	last->next = freed_logically;

	if(freed_logically != NULL) {
		freed_logically->prev = last;
	}

	first->prev = NULL;
	freed_logically = first;
}

inline void *wlpdstm::MemoryManager::Malloc(size_t size) {
#ifdef MM_TBB_MALLOC
	void *ret = scalable_malloc(size);
#elif defined MM_HOARD
	void *ret = hoard_malloc(size);
#elif defined MM_LEA
	void *ret = lea_malloc(size);
#else
	void *ret = ::malloc(size);
#endif

	return ret;
}

inline void wlpdstm::MemoryManager::Free(void *ptr) {
#ifdef MM_TBB_MALLOC
	scalable_free(ptr);
#elif defined MM_HOARD
	hoard_free(ptr);
#elif defined MM_LEA
	lea_free(ptr);
#else
	::free(ptr);
#endif
}

inline void *wlpdstm::MemoryManager::TxMalloc(size_t size) {
	// first allocate memory
	void *ret = Malloc(size);

	// remember allocation
	alloced_in_tx->ptrs[alloced_in_tx->next_ptr++] = ret;

	// make sure freed_in_tx is never full
	if(alloced_in_tx->IsFull()) {
		PtrBuffer *pbuf = GetCleanPtrBuffer();
		alloced_in_tx->next = pbuf;
		pbuf->prev = alloced_in_tx;
		alloced_in_tx = pbuf;
	}

	return ret;
}

inline void wlpdstm::MemoryManager::TxFree(void *ptr) {
	// remember deallocation
	freed_in_tx->ptrs[freed_in_tx->next_ptr++] = ptr;

	// make sure freed_in_tx is never full
	if(freed_in_tx->IsFull()) {
		PtrBuffer *pbuf = GetCleanPtrBuffer();
		freed_in_tx->next = pbuf;
		pbuf->prev = freed_in_tx;
		freed_in_tx = pbuf;
	}
}

inline void wlpdstm::MemoryManager::FreeDominated() {
	// find the first one to delete
	PtrBuffer *curr = freed_logically;

	while(curr != NULL) {
		if(curr->init_ts) {
			if(curr->ts < mm_ts) {
				break;
			}
		}

		curr = curr->next;
	}

	// delete all ptrs in dominated buffers and put buffers in
	// reusable list
	if(curr != NULL) {
		PtrBuffer *first = curr;
		PtrBuffer *last;

		if(first->prev == NULL) {
			freed_logically = NULL;
		} else {
			first->prev->next = NULL;
		}

		// all the rest are also dominated
		while(curr != NULL) {
			last = curr;
			FreePtrs(curr);
			curr = curr->next;
		}

		PutReusable(first, last);
	}
}

inline void wlpdstm::MemoryManager::FreeMemoryPtr(void *ptr) {
	Free(ptr);
}

inline void wlpdstm::MemoryManager::FreePtrs(PtrBuffer *buf) {
	unsigned end = buf->next_ptr;

	for(unsigned i = 0;i < end;i++) {
		FreeMemoryPtr(buf->ptrs[i]);
	}
}

///////////////////////////////////////
// memory manager implementation end //
///////////////////////////////////////


////////////////////////////////////
// timestamp implementation start //
////////////////////////////////////

inline wlpdstm::VectorTimestamp::VectorTimestamp(
		const VectorTimestamp &vt) {
	unsigned len = vt.length;

	for(unsigned i = 0;i < len;i++) {
		GetElement(i) = vt[i];
	}

	length = len;
}

inline void wlpdstm::VectorTimestamp::RegisterThread(unsigned tid) {
	unsigned current_length;

	while(true) {
		current_length = atomic_load_no_barrier(&length);

		if(current_length < tid + 1) {
			if(atomic_cas_no_barrier(&length, current_length, tid + 1)) {
				break;
			}
		} else {
			break;
		}
	}
}

inline void wlpdstm::VectorTimestamp::Init() {
	for(unsigned i = 0;i < MAX_THREADS;i++) {
		GetElement(i) = 0;
	}

	length = 0;
}

inline void wlpdstm::VectorTimestamp::Increment(unsigned tid) {
	GetElement(tid)++;
}

inline unsigned long wlpdstm::VectorTimestamp::operator[](unsigned idx) const {
	return ts[idx * wlpdstm::CACHE_LINE_SIZE];
}

inline unsigned long &wlpdstm::VectorTimestamp::GetElement(unsigned idx) {
	return ts[idx * wlpdstm::CACHE_LINE_SIZE];
}

inline bool wlpdstm::operator<(const VectorTimestamp &older,
		const VectorTimestamp &newer) {
	unsigned len = older.length;

	for(unsigned i = 0;i < len;i++) {
		unsigned ots = older[i];
		unsigned nts = newer[i];

		if(!(nts > ots || nts % 2 == 0)) {
			return false;
		}
	}

	return true;
}

//////////////////////////////////
// timestamp implementation end //
//////////////////////////////////

#endif
