/**
 * @author Aleksandar Dragojevic aleksandar.dragojevic@epfl.ch
 */

#include <cstdlib>

#include "aborted_exc.h"
#include "common/version_lock.h"
#include "constants.h"
#include "common/log.h"
#include "common/timestamp.h"
#include "memory.h"
#include "tid.h"
#include "common/tls.h"

#ifdef USE_PREEMPTIVE_WAITING
#include "common/preemptive_utils.h"
#endif // USE_PREEMPTIVE_WAITING

#define ADDRESS_SPACE_SIZE 32
#define LOCK_EXTENT_SIZE 6

// Version lock table covers 32 bit address space in chunks of 64 bytes.
#define VERSION_LOCK_TABLE_SIZE (2 << (ADDRESS_SPACE_SIZE - LOCK_EXTENT_SIZE))

#define CM_PHASE_THRESHOLD 1

namespace wlpdstm {

	class TxMixinvIR {

		//////////////////////////////////
		// log related structures start //
		//////////////////////////////////

		struct ReadLogEntry {
			VersionLock *address;
			VersionLock version;
		};

		struct WriteWordLogEntry {
			void *address;
			void *value;
			WriteWordLogEntry *next;
		};

		struct WriteWordLocalLogEntry {
			void *address;
			void *value;
		};

		struct WriteLogEntry {
			WriteLogEntry() : entry_version(0) {  }

			VersionLock *address;

			// needed to allow concurrent read/writes
			VersionLock old_version;

			TxMixinvIR *owner;

			WriteWordLogEntry *head;

			unsigned entry_version;

			// TODO: perhaps add a bitmap to speed up lookups
			// might not be needed as memory regions are short

			// methods
			void InsertWordLogEntry(void *address, void *value);

			WriteWordLogEntry *FindWordLogEntry(void *address);
		};

		typedef Log<ReadLogEntry> ReadLog;
		typedef Log<WriteLogEntry> WriteLog;
		typedef Log<WriteWordLocalLogEntry> WriteLocalLog;
		typedef Log<WriteWordLogEntry> WriteWordLogMemPool;

		////////////////////////////////
		// log related structures end //
		////////////////////////////////

		// transaction state
		public:
			enum TxState {
				TX_STATE_IDLE,
				TX_STATE_RUNNING,
				TX_STATE_COMMITTING
			};

		public:
			static void GlobalInit();

			void ThreadInit();

			void TxStart();

			void TxCommit();

			void TxAbort();

			void TxAbortFull();

			void WriteWord(void *address, void *val);

			void WriteWordLocal(void *address, void *val);

			void *ReadWord(void *address);

			void *TxMalloc(size_t size);

			void TxFree(void *ptr, size_t size);

			void TxRestart();

		private:
			void LockMemoryBlock(void *address, size_t size);

			static VersionLock *map_address_to_lock(void *address);

			WriteLogEntry *LockMemoryStripe(VersionLock *lock_address);

			void Rollback();

			bool Validate();

			bool Extend();

			bool LockedByMe(WriteLogEntry *log_entry);

			///////////////////////////
			// contention management //
			///////////////////////////

			bool ShouldAbortWrite(VersionLock *lock_address);

			bool CMStrongerThan(TxMixinvIR *other);

			bool Retrying() const;

			bool SecondCMPhase() const;

		private:
			ReadLog read_log;

			WriteLog write_log;

			WriteLocalLog write_local_log;

			WriteWordLogMemPool write_word_log_mem_pool;

			unsigned valid_ts;

			unsigned succ_abort_cnt;

			TxState state;

			unsigned pad[4];

			////////////////////////////////
			// contention management data //
			////////////////////////////////

			unsigned greedy_ts;

			bool aborted_externally;

			//////////////////////
			// shared variables //
			//////////////////////

			static VersionLock version_lock_table[VERSION_LOCK_TABLE_SIZE];

			static GlobalTimestamp commit_ts;

			static GlobalTimestamp cm_ts;

			///////////////
			// constants //
			///////////////

			static const VersionLock MINIMUM_VERSION;

		public:
			Tid tid;

			MemoryManager mm;

	} __attribute__ ((aligned(1024)));
}

inline void wlpdstm::TxMixinvIR::GlobalInit() {
	// initialize object locks
	VersionLock initial_version = get_version_lock(MINIMUM_TS + 1);

	// initialize global table lock
	for(int i = 0;i < VERSION_LOCK_TABLE_SIZE;i++) {
		version_lock_table[i] = initial_version;
	}

	// initialize memory manager
	MemoryManager::GlobalInit();
}

// TODO: the most natural way would be to map addresses to lock based
// on objects that addresses point to, but this is not so simple
inline wlpdstm::VersionLock *wlpdstm::TxMixinvIR::map_address_to_lock(
		void *address) {
	return version_lock_table + ((unsigned)address >> LOCK_EXTENT_SIZE);
}

inline void wlpdstm::TxMixinvIR::ThreadInit() {
	aborted_externally = false ;
	succ_abort_cnt = 0;
	state = TX_STATE_IDLE;

	// locally initialize memory manager
	mm.ThreadInit(tid.Get());
}

inline void wlpdstm::TxMixinvIR::TxStart() {
	mm.TxStart();

	// get validity timestamp
	valid_ts = commit_ts.readCurrentTs();

	// decide which cm phase to use
	if(SecondCMPhase()) {
		greedy_ts = cm_ts.getNextTs() + 1;
	} else if(!Retrying()) {
		greedy_ts = MINIMUM_TS;
	}

	// reset aborted flag
	aborted_externally = false;

	// start running
	atomic_store_full(&state, TX_STATE_RUNNING);
}

inline void wlpdstm::TxMixinvIR::TxCommit() {
	if(!write_log.empty()) {
		// first change state
		state = TX_STATE_COMMITTING;

		// increase global timestamp
		unsigned ts = commit_ts.getNextTsFull() + 1;

		if(ts != valid_ts + 1 && !Validate()) {
			TxRestart();
		}

		// generate new version in correct format
		VersionLock commitVersion = get_version_lock(ts);

		// process redo log
		for(WriteLog::iterator iter = write_log.begin();iter.hasNext();iter.next()) {
			WriteLogEntry &entry = *iter;
			entry.old_version = MINIMUM_VERSION;
		}

		atomic_barrier_full();

		for(WriteLog::iterator iter = write_log.begin();iter.hasNext();iter.next()) {
			WriteLogEntry &entry = *iter;

			// now update actual values
			WriteWordLogEntry *word_log_entry = entry.head;

			while(word_log_entry != NULL) {
				*(void **)word_log_entry->address =
					word_log_entry->value;
				word_log_entry = word_log_entry->next;
			}

			// release the lock
			atomic_store_release(entry.address, commitVersion);
		}
	}

	// finish committing values
	state = TX_STATE_IDLE;

	write_log.clear();
	read_log.clear();
	write_local_log.clear();
	write_word_log_mem_pool.clear();

	succ_abort_cnt = 0;

	// commit mem
	mm.TxCommit();
}

inline void wlpdstm::TxMixinvIR::Rollback() {
	// rollback local writes
	for(WriteLocalLog::iterator iter = write_local_log.begin();iter.hasNext();iter.next()) {
		WriteWordLocalLogEntry &entry = *iter;
		*(void **)entry.address = entry.value;
	}

	// drop locks
	for(WriteLog::iterator iter = write_log.begin();iter.hasNext();iter.next()) {
		WriteLogEntry &entry = *iter;
		atomic_store_no_barrier(entry.address, entry.old_version);
	}

	// empty logs
	read_log.clear();
	write_log.clear();
	write_local_log.clear();
	write_word_log_mem_pool.clear();

#ifdef USE_PREEMPTIVE_WAITING
	pre_yield();
#endif

	mm.TxAbort();
}

inline wlpdstm::TxMixinvIR::WriteLogEntry *wlpdstm::TxMixinvIR::LockMemoryStripe(
		VersionLock *lock_address) {
	// read lock value
	VersionLock version = (VersionLock)atomic_load_no_barrier(lock_address);

	if(is_locked(version)) {
		WriteLogEntry *log_entry = (WriteLogEntry *)get_ptr(version);

		if(LockedByMe(log_entry)) {
			return log_entry;
		}
	}

	while(true) {
		if(is_locked(version)) {
			if(ShouldAbortWrite(lock_address)) {
				TxRestart();
			} else {
				version = (VersionLock)atomic_load_acquire(lock_address);
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}
		}

		// the word read was unlocked, check validity
		// TODO: Validity need not be checked if the word was not read
		// before, but this requires a pass over read set or use of
		// some lookup mechanism (Bloom filter being used in other
		// implementations), for now skip this check and assume all
		// written words were read before.
		//
		// I will implement something later and check if it helps or not.
		if(get_value(version) > valid_ts) {
			// TODO: check if extending really helps or just slows down
			// unnecessarily
			if(!Extend()) {
				// must abort here
				TxRestart();
			}
		}

		// prepare write log entry
		WriteLogEntry *log_entry = write_log.get_next();
		log_entry->address = lock_address;
		log_entry->old_version = version;
		log_entry->head = NULL; // need this here
		log_entry->owner = this;
		++log_entry->entry_version;

		// get lock value
		VersionLock locked_entry = get_locked(log_entry);

		// try to cas log_entry into
		if(atomic_cas_release(lock_address, version, locked_entry)) {
			// success
			return log_entry;
		} else {
			// return last element to the log
			write_log.delete_last();

			// read version again
			version = (VersionLock)atomic_load_acquire(lock_address);
#ifdef USE_PREEMPTIVE_WAITING
			pre_yield();
#endif
		}
	}

	// this can never happen
	return NULL;
}

inline void wlpdstm::TxMixinvIR::WriteWord(void *address, void *value) {
	// map address to the lock
	VersionLock *lock_address = map_address_to_lock(address);

	// try to lock the address
	WriteLogEntry *log_entry = LockMemoryStripe(lock_address);

	// insert (address, value) pair into the log
	log_entry->InsertWordLogEntry(address, value);
}

inline void wlpdstm::TxMixinvIR::WriteWordLocal(void *address, void *val) {
	WriteWordLocalLogEntry *log_entry = write_local_log.get_next();

	log_entry->address = address;
	log_entry->value = *(void **)address;

	// update location
	*(void **)address = val;
}

inline wlpdstm::TxMixinvIR::WriteWordLogEntry *wlpdstm::TxMixinvIR::WriteLogEntry::FindWordLogEntry(
		void *address) {
	WriteWordLogEntry *curr = head;

	while(curr != NULL) {
		if(curr->address == address) {
			break;
		}

		curr = curr->next;
	}

	return curr;
}

inline void wlpdstm::TxMixinvIR::WriteLogEntry::InsertWordLogEntry(
		void *address, void *value) {
	WriteWordLogEntry *entry = FindWordLogEntry(address);

	if(entry == NULL) {
		entry = owner->write_word_log_mem_pool.get_next();
		entry->address = address;
		entry->next = head;
		head = entry;
	}

	entry->value = value;
}

inline void *wlpdstm::TxMixinvIR::ReadWord(void *address) {
	// map address to the lock
	VersionLock *lock_address = map_address_to_lock(address);

	// read lock value
	VersionLock version = (VersionLock)atomic_load_no_barrier(lock_address);

	// if locked by me, return value
	if(is_locked(version)) {
		WriteLogEntry *log_entry = (WriteLogEntry *)get_ptr(version);

		if(LockedByMe(log_entry)) {
			WriteWordLogEntry *word_log_entry = log_entry->FindWordLogEntry(address);

			if(word_log_entry != NULL) {
				// if it was written return from log
				return word_log_entry->value;
			} else {
				// if it was not written return from memory
				return (void *)atomic_load_no_barrier(address);
			}

			// no need to add this address to read set, because it cannot
			// be changed as it is locked by this tx
		}
	}

	VersionLock ver;
	void *value;

	while(true) {
		bool locked = is_locked(version);

		if(!locked) {
			value = (void *)atomic_load_acquire(address);
			VersionLock version_2 = (VersionLock)atomic_load_acquire(lock_address);

			if(version != version_2) {
				version = version_2;
	#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
	#endif
				continue;
			}

			ver = version;
			break;
		} else {
			// it is locked
			WriteLogEntry *log_entry = (WriteLogEntry *)get_ptr(version);

			// check owner state for committing owner
			TxMixinvIR *owner = log_entry->owner;
			TxState owner_state = owner->state;

			if(owner_state == TX_STATE_COMMITTING) {
				version = (VersionLock)atomic_load_acquire(lock_address);
				// TODO: wait while it finishes here perhaps
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}

			unsigned entry_version = (unsigned)atomic_load_acquire(
				&log_entry->entry_version);

			value = (void *)atomic_load_acquire(address);

			VersionLock old_version = (VersionLock)atomic_load_acquire(
				&log_entry->old_version);

			unsigned entry_version_2 = (unsigned)atomic_load_acquire(
				&log_entry->entry_version);

			// if owner got to committing in the meantime try again
			// if the old owner finished committing and was able to
			// acquire the object and cas in the old log_entry try again
			// NOTE: there is a minor possibility that the owner could
			// run the whole entry_version around to the exact same number
			// and cause inconsistencies
			if(entry_version != entry_version_2 ||
					old_version == MINIMUM_VERSION) {
				version = (VersionLock)atomic_load_acquire(lock_address);
#ifdef USE_PREEMPTIVE_WAITING
				pre_yield();
#endif
				continue;
			}

			ver = old_version;
			break;
		}
	}

	// check validity
	if(get_value(ver) > valid_ts) {
		if(!Extend()) {
			// must abort here
			TxRestart();
		}
	}

	// here all is consistent and valid :)
	ReadLogEntry *entry = read_log.get_next();
	entry->address = lock_address;
	entry->version = ver;
	return value;
}

inline bool wlpdstm::TxMixinvIR::Validate() {
	ReadLog::iterator iter;

	for(iter = read_log.begin();iter.hasNext();iter.next()) {
		ReadLogEntry &entry = *iter;

		VersionLock currentVersion = (VersionLock)atomic_load_no_barrier(
			entry.address);

		if(!is_locked(currentVersion)) {
			if(currentVersion != entry.version) {
				return false;
			}
		} else {
			WriteLogEntry *log_entry = (WriteLogEntry *)get_ptr(currentVersion);

			if(LockedByMe(log_entry)) {
				continue;
			}

			while(true) {
				if(!is_locked(currentVersion)) {
					if(currentVersion != entry.version) {
						return false;
					}
				} else {
					log_entry = (WriteLogEntry *)get_ptr(currentVersion);
					TxMixinvIR *owner = (TxMixinvIR *)log_entry->owner;
					TxState owner_state = (TxState)atomic_load_acquire(&owner->state);

					if(owner_state == TX_STATE_COMMITTING) {
						return false;
					}

					VersionLock old_version = (VersionLock)atomic_load_acquire(
						&log_entry->old_version);

					if(old_version != entry.version) {
						return false;
					}

					// reread version here to avoid getting the same
					// log entry, but for a different stripe
					VersionLock version_2 = (VersionLock)atomic_load_acquire(
						entry.address);

					// if version changed in the meantime re-read
					if(currentVersion != version_2) {
						currentVersion = version_2;
						continue;
					}
				}

				break;
			}
		}
	}

	return true;
}

inline bool wlpdstm::TxMixinvIR::Extend() {
	unsigned ts = commit_ts.readCurrentTs();

	if(Validate()) {
		valid_ts = ts;
		return true;
	}

	return false;
}

inline void wlpdstm::TxMixinvIR::LockMemoryBlock(void *address, size_t size) {
	// calc last address of the block
	void *end_address = (char *)address + size - 1;

	// get start and end version lock address
	// knowledge about how addresses are mapped to locks is used here
	VersionLock *start = map_address_to_lock(address);
	VersionLock *end = map_address_to_lock(end_address);

	// iterate throiugh locks and lock them all
	for(VersionLock *iter = start;iter <= end;iter++) {
		LockMemoryStripe(iter);
	}
}

inline void *wlpdstm::TxMixinvIR::TxMalloc(size_t size) {
	return mm.TxMalloc(size);
}

inline void wlpdstm::TxMixinvIR::TxFree(void *ptr, size_t size) {
	LockMemoryBlock(ptr, size);
	mm.TxFree(ptr);
}

inline bool wlpdstm::TxMixinvIR::LockedByMe(WriteLogEntry *log_entry) {
	return write_log.contains(log_entry);
}

inline void wlpdstm::TxMixinvIR::TxRestart() {
	Rollback();

	state = TX_STATE_IDLE;
	succ_abort_cnt++;

	throw RestartedException();
}

inline void wlpdstm::TxMixinvIR::TxAbort() {
	Rollback();
	
	state = TX_STATE_IDLE;
	succ_abort_cnt = 0;

	throw AbortedException();
}


inline bool wlpdstm::TxMixinvIR::ShouldAbortWrite(
		VersionLock *lock_address) {
	if(aborted_externally) {
		return true;
	}

	if(greedy_ts == MINIMUM_TS) {
		return true;
	}

	VersionLock version = (VersionLock)atomic_load_no_barrier(lock_address);

	if(is_locked(version)) {
		WriteLogEntry *log_entry = (WriteLogEntry *)get_ptr(version);
		TxMixinvIR *owner = log_entry->owner;

		if(CMStrongerThan(owner)) {
			if(!owner->aborted_externally) {
				owner->aborted_externally = true;
			}

			return false;
		}

		return true;
	}

	return false;
}

inline bool wlpdstm::TxMixinvIR::CMStrongerThan(TxMixinvIR *other) {
	unsigned other_ts = other->greedy_ts;

	if(greedy_ts == MINIMUM_TS) {
		return false;
	}

	return other_ts == MINIMUM_TS || (greedy_ts < other_ts);
}

inline bool wlpdstm::TxMixinvIR::Retrying() const {
	return succ_abort_cnt != 0;
}

inline bool wlpdstm::TxMixinvIR::SecondCMPhase() const {
	return succ_abort_cnt == CM_PHASE_THRESHOLD;
}
